import numpy as np
import matplotlib.pyplot as plt
from matplotlib import widgets

KEEP_ALIVE_ARR = []
def keep_alive(x):
    KEEP_ALIVE_ARR.append(x)
    return x

def augment_ones(x):
    return np.concatenate((np.ones([x.shape[0]]+([1]*(x.ndim-1))), x), axis=-1)

point = None
def gen_do_step(ax3d, fig, ax3d_title, theta0, theta1, theta, f, df, ddf):
    def do_step(event):
        ax3d.clear()
        ax3d.plot_surface(theta0, theta1, f(theta), color=(0., 0., 1., 0.5))
        if event is not None:
            global point
            if point is None:
                point = point0
            else:
                point -= np.linalg.inv(ddf(point)) @ df(point)
            ax3d.plot(point[0], point[1], f(point), 'go', markersize=6)
        ax3d.set_xlabel('$b$')
        ax3d.set_ylabel('$w$')
        ax3d.set_title(ax3d_title)
        fig.canvas.draw_idle()
    return do_step

def build_figure(X_pre, y, lamb, theta0_lim, theta1_lim, fig, ax2d_title, ax3d_title):
    def f(theta):
        raw = -((theta @ X.T @ y) - np.sum(np.log(1 + np.exp(theta @ X.T)), axis=-1))
        regul = lamb * (np.linalg.norm(theta, axis=-1)**2)
        return raw + regul
    def df(theta):
        tmp = y - (1/(1 + np.exp(-theta @ X.T)))
        raw = -(tmp @ X)
        regul = 2 * lamb * theta
        return raw + regul
    def ddf(theta):
        e = np.exp(-theta @ X.T)
        s = -e/((1+e)**2)
        raw = -np.einsum('...n,nd,ne->...de', s, X, X)
        regul = 2 * lamb * np.eye(X.shape[-1])
        return raw + regul
    X = augment_ones(X_pre)
    N = 128
    theta0, theta1 = np.meshgrid(
        np.linspace(*theta0_lim, N),
        np.linspace(*theta1_lim, N)
    )
    theta = np.stack((theta0, theta1), axis=-1)
    fig_arr = fig.subfigures(2, 1, height_ratios=[0.1, 0.9])
    ax2d, ax_btn = fig_arr[0].subplots(1, 2, width_ratios=[0.8, 0.2])
    mask = (1 == y)
    ax2d.plot(np.squeeze(X_pre, -1)[~mask], np.zeros(X_pre.shape[0])[~mask], 'ro')
    ax2d.plot(np.squeeze(X_pre, -1)[mask], np.zeros(X_pre.shape[0])[mask], 'bo')
    ax2d.spines['left'].set_color('none')
    ax2d.spines['bottom'].set_position('zero')
    ax2d.spines['right'].set_color('none')
    ax2d.spines['top'].set_color('none')
    ax2d.set_xticks(np.squeeze(X_pre, -1))
    ax2d.set_yticks([])
    fig_arr[0].suptitle(ax2d_title)
    ax3d = fig_arr[1].add_subplot(projection='3d')
    do_step = gen_do_step(ax3d, fig, ax3d_title, theta0, theta1, theta, f, df, ddf)
    btn = keep_alive(widgets.Button(ax_btn, 'Newton Step'))
    btn.on_clicked(do_step)
    do_step(None)
    ax3d.view_init(azim=118, elev=30)

point0 = None
def main(idx):
    global point0
    if 0 == idx:
        X_pre = np.array([[-1], [1], [3]])
        y = np.array([1, 0, 1])
        theta0_lim = [-5, 5]
        theta1_lim = [-5, 5]
        fig = plt.figure()
        ax2d_title = 'Non-Separable Dataset'
        ax3d_title = 'Logistic Regression Cost'
        lamb = 0.
        point0 = np.zeros(2)
    elif 1 == idx:
        X_pre = np.array([[-1], [1], [3]])
        y = np.array([0, 0, 1])
        theta0_lim = [-5, 5]
        theta1_lim = [-5, 5]
        fig = plt.figure()
        ax2d_title = 'Separable Dataset'
        ax3d_title = 'Logistic Regression Cost'
        lamb = 0.
        point0 = np.zeros(2)
    elif 2 == idx:
        X_pre = np.array([[-1], [1], [3]])
        y = np.array([0, 0, 1])
        theta0_lim = [-20, 20]
        theta1_lim = [-20, 20]
        fig = plt.figure()
        ax2d_title = 'Separable Dataset'
        ax3d_title = 'Logistic Regression Cost'
        lamb = 0.
        point0 = np.array([1., 1.])
    elif 3 == idx:
        X_pre = np.array([[-1], [1], [3]])
        y = np.array([0, 0, 1])
        theta0_lim = [-20, 20]
        theta1_lim = [-20, 20]
        fig = plt.figure()
        ax2d_title = 'Separable Dataset'
        ax3d_title = '(Regularized) Logistic Regression Cost'
        lamb = 1e-1
        point0 = np.array([1., 1.])
    else:
        exit(1)
    build_figure(X_pre, y, lamb, theta0_lim, theta1_lim, fig, ax2d_title, ax3d_title)
    plt.show()

import sys
main(int(sys.argv[1]))
